"""Defines a logger which logs JSON lines to a file."""

import json
import sys
from typing import Any, Literal, Mapping, TextIO

from jaxtyping import Array

from xax.task.logger import (
    LogError,
    LogErrorSummary,
    LoggerImpl,
    LogLine,
    LogPing,
    LogScalar,
    LogStatus,
    LogString,
)


def get_json_value(value: Any) -> Any:  # noqa: ANN401
    if isinstance(value, Array):
        value = value.item()
    return value


class JsonLogger(LoggerImpl):
    def __init__(
        self,
        log_stream: TextIO = sys.stdout,
        err_log_stream: TextIO = sys.stderr,
        flush_immediately: bool = False,
        open_mode: Literal["w", "a"] = "w",
        line_sep: str = "\n",
        remove_unicode_from_namespaces: bool = True,
        log_interval_seconds: float = 10.0,
    ) -> None:
        """Defines a simpler logger which logs to stdout.

        Args:
            log_stream: The stream to log to.
            err_log_stream: The stream to log errors to.
            flush_immediately: Whether to flush the file after every write.
            open_mode: The file open mode.
            line_sep: The line separator to use.
            remove_unicode_from_namespaces: Whether to remove unicode from
                namespaces. This is the typical behavior for namespaces that
                use ASCII art for visibility in other logs, but in the JSON
                log file should be ignored.
            log_interval_seconds: The interval between successive log lines.
        """
        super().__init__(log_interval_seconds)

        self.log_stream = log_stream
        self.err_log_stream = err_log_stream
        self.flush_immediately = flush_immediately
        self.open_mode = open_mode
        self.line_sep = line_sep
        self.remove_unicode_from_namespaces = remove_unicode_from_namespaces

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    @property
    def fp(self) -> TextIO:
        return self.log_stream

    @property
    def err_fp(self) -> TextIO:
        return self.err_log_stream

    def get_json(self, line: LogLine) -> str:
        data: dict = {"state": line.state.to_dict()}

        def add_logs(log: Mapping[str, Mapping[str, LogScalar | LogString]], data: dict) -> None:
            for namespace, values in log.items():
                if self.remove_unicode_from_namespaces:
                    namespace = namespace.encode("ascii", errors="ignore").decode("ascii").strip()
                if namespace not in data:
                    data[namespace] = {}
                for k, v in values.items():
                    data[namespace][k] = get_json_value(v.value)

        add_logs(line.scalars, data)
        add_logs(line.strings, data)
        return json.dumps(data)

    def write(self, line: LogLine) -> None:
        self.fp.write(self.get_json(line))
        self.fp.write(self.line_sep)
        if self.flush_immediately:
            self.fp.flush()

    def write_error_summary(self, error_summary: LogErrorSummary) -> None:
        pass

    def log_file(self, name: str, contents: str) -> None:
        pass

    def write_error(self, error: LogError) -> None:
        self.err_fp.write(error.message)
        if error.location is not None:
            self.err_fp.write(f" ({error.location})")
        self.err_fp.write(self.line_sep)
        if self.flush_immediately:
            self.err_fp.flush()

    def write_ping(self, ping: LogPing) -> None:
        self.fp.write(
            json.dumps(
                {
                    "ping": {
                        "message": ping.message,
                        "filename": ping.filename,
                        "lineno": ping.lineno,
                    },
                },
            ),
        )
        self.fp.write(self.line_sep)
        if self.flush_immediately:
            self.fp.flush()

    def write_status(self, status: LogStatus) -> None:
        self.fp.write(
            json.dumps(
                {
                    "status": {
                        "message": status.message,
                        "filename": status.filename,
                        "lineno": status.lineno,
                    },
                },
            ),
        )
        self.fp.write(self.line_sep)
        if self.flush_immediately:
            self.fp.flush()



"""Defines a single interface for all the mixins."""

from xax.task.mixins.artifacts import ArtifactsConfig, ArtifactsMixin
from xax.task.mixins.checkpointing import CheckpointingConfig, CheckpointingMixin
from xax.task.mixins.compile import CompileConfig, CompileMixin
from xax.task.mixins.cpu_stats import CPUStatsConfig, CPUStatsMixin
from xax.task.mixins.data_loader import DataloadersConfig, DataloadersMixin
from xax.task.mixins.gpu_stats import GPUStatsConfig, GPUStatsMixin
from xax.task.mixins.logger import LoggerConfig, LoggerMixin
from xax.task.mixins.process import ProcessConfig, ProcessMixin
from xax.task.mixins.runnable import RunnableConfig, RunnableMixin
from xax.task.mixins.step_wrapper import StepContextConfig, StepContextMixin
from xax.task.mixins.train import TrainConfig, TrainMixin



"""Defines some probability distribution helper functions.

In general, it is preferrable to use Distrax or another library, but we wanted
to have a simple interface of our own so that we can quickly upgrade Jax
versions (since Distrax is tied pretty closely to Tensorflow).
"""

__all__ = [
    "Distribution",
    "Categorical",
    "Normal",
    "MixtureOfGaussians",
]

from abc import ABC, abstractmethod

import jax
import jax.numpy as jnp
from jaxtyping import Array, PRNGKeyArray

STD_CLIP = 1e-6
LOGIT_CLIP = 6.0


class Distribution(ABC):
    @abstractmethod
    def log_prob(self, x: Array) -> Array: ...

    @abstractmethod
    def sample(self, key: PRNGKeyArray) -> Array: ...

    @abstractmethod
    def mode(self) -> Array: ...

    @abstractmethod
    def entropy(self) -> Array: ...


class Categorical(Distribution):
    def __init__(self, logits_nc: Array, logit_clip: float = LOGIT_CLIP) -> None:
        """Initialize a categorical distribution.

        Args:
            logits_nc: Array of shape (..., n_categories) containing logits
            logit_clip: Clipping value for logits
        """
        self.logits_nc = jnp.clip(logits_nc, -logit_clip, logit_clip)

    @property
    def num_categories(self) -> int:
        return self.logits_nc.shape[-1]

    def log_prob(self, x_n: Array) -> Array:
        log_probs_n = jax.nn.log_softmax(self.logits_nc, axis=-1)
        return log_probs_n[x_n]

    def sample(self, key: PRNGKeyArray) -> Array:
        return jax.random.categorical(key, self.logits_nc, axis=-1)

    def mode(self) -> Array:
        return self.logits_nc.argmax(axis=-1)

    def entropy(self) -> Array:
        probs = jax.nn.softmax(self.logits_nc, axis=-1)
        log_probs = jax.nn.log_softmax(self.logits_nc, axis=-1)
        return -jnp.sum(probs * log_probs, axis=-1)


class Normal(Distribution):
    def __init__(self, loc_n: Array, scale_n: Array, std_clip: float = STD_CLIP) -> None:
        """Initialize a normal distribution.

        Args:
            loc_n: Mean of the distribution
            scale_n: Standard deviation of the distribution
            std_clip: Minimum standard deviation
        """
        self.loc_n = loc_n
        self.scale_n = jnp.clip(scale_n, min=std_clip)

    def log_prob(self, x: Array) -> Array:
        return -0.5 * jnp.log(2 * jnp.pi) - jnp.log(self.scale_n) - (x - self.loc_n) ** 2 / (2 * self.scale_n**2)

    def sample(self, key: PRNGKeyArray) -> Array:
        return self.loc_n + self.scale_n * jax.random.normal(key, self.loc_n.shape)

    def mode(self) -> Array:
        return self.loc_n

    def entropy(self) -> Array:
        return jnp.log(2 * jnp.pi * jnp.e) + jnp.log(self.scale_n)


class MixtureOfGaussians(Distribution):
    def __init__(
        self,
        means_nm: Array,
        stds_nm: Array,
        logits_nm: Array,
        std_clip: float = STD_CLIP,
        logit_clip: float = LOGIT_CLIP,
    ) -> None:
        """Initialize a mixture of Gaussians.

        Args:
            means_nm: Array of shape (..., n_components) containing means
            stds_nm: Array of shape (..., n_components) containing standard deviations
            logits_nm: Array of shape (..., n_components) containing mixing logits
            std_clip: Minimum standard deviation
            logit_clip: Clipping value for logits
        """
        self.means_nm = means_nm
        self.stds_nm = jnp.clip(stds_nm, min=std_clip)
        self.logits_nm = jnp.clip(logits_nm, -logit_clip, logit_clip)

    def log_prob(self, x_n: Array) -> Array:
        # Expand x to match component dimensions
        x_n_expanded = x_n[..., None]  # Shape: (..., 1)

        # Compute log probabilities for each component
        component_log_probs = (
            -0.5 * jnp.log(2 * jnp.pi)
            - jnp.log(self.stds_nm)
            - (x_n_expanded - self.means_nm) ** 2 / (2 * self.stds_nm**2)
        )

        # Compute mixing weights
        mixing_logits = jax.nn.log_softmax(self.logits_nm, axis=-1)

        # Combine using log-sum-exp trick for numerical stability
        return jax.scipy.special.logsumexp(component_log_probs + mixing_logits, axis=-1)

    def sample(self, key: PRNGKeyArray) -> Array:  # Sample component indices
        component_key, sample_key = jax.random.split(key)
        component_indices = jax.random.categorical(component_key, self.logits_nm, axis=-1)

        # Sample from selected components using advanced indexing
        # We need to handle the case where we have batch dimensions
        batch_shape = self.means_nm.shape[:-1]  # All dimensions except the last (components)

        # Reshape for easier indexing
        means_flat = self.means_nm.reshape(-1, self.means_nm.shape[-1])
        stds_flat = self.stds_nm.reshape(-1, self.stds_nm.shape[-1])
        indices_flat = component_indices.reshape(-1)

        # Get selected means and stds
        selected_means = means_flat[jnp.arange(len(indices_flat)), indices_flat]
        selected_stds = stds_flat[jnp.arange(len(indices_flat)), indices_flat]

        # Generate random noise
        noise = jax.random.normal(sample_key, selected_means.shape)

        # Reshape back to original batch shape
        samples_n = selected_means + selected_stds * noise
        return samples_n.reshape(batch_shape)

    def mode(self) -> Array:
        """Return the mode of the mixture (approximate - returns mean of highest weight component)."""
        mixing_weights = jax.nn.softmax(self.logits_nm, axis=-1)
        max_weight_idx = jnp.argmax(mixing_weights, axis=-1)

        # Use advanced indexing to get the means of the highest weight components
        batch_shape = self.means_nm.shape[:-1]
        means_flat = self.means_nm.reshape(-1, self.means_nm.shape[-1])
        indices_flat = max_weight_idx.reshape(-1)

        selected_means = means_flat[jnp.arange(len(indices_flat)), indices_flat]
        return selected_means.reshape(batch_shape)

    def entropy(self) -> Array:
        """Compute entropy of the mixture (approximate)."""
        mixing_weights = jax.nn.softmax(self.logits_nm, axis=-1)
        component_entropies = jnp.log(2 * jnp.pi * jnp.e) + jnp.log(self.stds_nm)

        # Weighted sum of component entropies plus mixing entropy
        weighted_entropies = jnp.sum(mixing_weights * component_entropies, axis=-1)
        mixing_entropy = -jnp.sum(mixing_weights * jnp.log(mixing_weights + 1e-8), axis=-1)

        return weighted_entropies + mixing_entropy



#!/usr/bin/env python
"""Setup script for the project."""

import re

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description: str = f.read()

with open("xax/requirements.txt", "r", encoding="utf-8") as f:
    requirements: list[str] = f.read().splitlines()

with open("xax/requirements-dev.txt", "r", encoding="utf-8") as f:
    requirements_dev: list[str] = f.read().splitlines()

with open("xax/__init__.py", "r", encoding="utf-8") as fh:
    version_re = re.search(r"^__version__ = \"([^\"]*)\"", fh.read(), re.MULTILINE)
assert version_re is not None, "Could not find version in xax/__init__.py"
version: str = version_re.group(1)


setup(
    name="xax",
    version=version,
    description="A library for fast Jax experimentation",
    author="Benjamin Bolte",
    url="https://github.com/kscalelabs/xax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.11",
    install_requires=requirements,
    tests_require=requirements_dev,
    extras_require={
        "dev": requirements_dev,
    },
    package_data={
        "xax": [
            "py.typed",
            "requirements*.txt",
        ],
    },
    entry_points={
        "console_scripts": [
            "xax-edit-config=xax.cli.edit_config:main",
        ],
    },
)



"""Defines functions for parallelism."""

import os

_RANK: int | None = None
_LOCAL_RANK: int | None = None
_WORLD_SIZE: int | None = None
_LOCAL_WORLD_SIZE: int | None = None
_MASTER_ADDR: str | None = None
_MASTER_PORT: int | None = None
_INIT_METHOD: str | None = None


def set_rank(rank: int) -> None:
    global _RANK

    if rank != _RANK:
        _RANK = rank
        os.environ["RANK"] = str(rank)
    else:
        raise ValueError(f"Rank {rank} is already set")


def get_rank_optional() -> int | None:
    return _RANK


def get_rank() -> int:
    return 0 if _RANK is None else _RANK


def clear_rank() -> None:
    global _RANK

    _RANK = None
    os.environ.pop("RANK", None)


def set_local_rank(rank: int) -> None:
    global _LOCAL_RANK

    if rank != _LOCAL_RANK:
        _LOCAL_RANK = rank
        os.environ["LOCAL_RANK"] = str(rank)
    else:
        raise ValueError(f"Local rank {rank} is already set")


def get_local_rank_optional() -> int | None:
    return _LOCAL_RANK


def get_local_rank() -> int:
    return 0 if _LOCAL_RANK is None else _LOCAL_RANK


def clear_local_rank() -> None:
    global _LOCAL_RANK

    _LOCAL_RANK = None
    os.environ.pop("LOCAL_RANK", None)


def set_world_size(world_size: int) -> None:
    global _WORLD_SIZE

    if world_size != _WORLD_SIZE:
        _WORLD_SIZE = world_size
        os.environ["WORLD_SIZE"] = str(world_size)
    else:
        raise ValueError(f"World size {world_size} is already set")


def get_world_size_optional() -> int | None:
    return _WORLD_SIZE


def get_world_size() -> int:
    return 1 if _WORLD_SIZE is None else _WORLD_SIZE


def clear_world_size() -> None:
    global _WORLD_SIZE

    _WORLD_SIZE = None
    os.environ.pop("WORLD_SIZE", None)


def set_local_world_size(local_world_size: int) -> None:
    global _LOCAL_WORLD_SIZE

    if local_world_size != _LOCAL_WORLD_SIZE:
        _LOCAL_WORLD_SIZE = local_world_size
        os.environ["LOCAL_WORLD_SIZE"] = str(local_world_size)
    else:
        raise ValueError(f"World size {local_world_size} is already set")


def get_local_world_size_optional() -> int | None:
    return _LOCAL_WORLD_SIZE


def get_local_world_size() -> int:
    return 1 if _LOCAL_WORLD_SIZE is None else _LOCAL_WORLD_SIZE


def clear_local_world_size() -> None:
    global _LOCAL_WORLD_SIZE

    _LOCAL_WORLD_SIZE = None
    os.environ.pop("LOCAL_WORLD_SIZE", None)


def set_master_addr(master_addr: str) -> None:
    global _MASTER_ADDR

    if master_addr != _MASTER_ADDR:
        os.environ["MASTER_ADDR"] = _MASTER_ADDR = master_addr
    else:
        raise ValueError(f"Master address {master_addr} is already set")


def get_master_addr() -> str:
    assert _MASTER_ADDR is not None, "Master address is not yet set"
    return _MASTER_ADDR


def clear_master_addr() -> None:
    global _MASTER_ADDR

    _MASTER_ADDR = None
    os.environ.pop("MASTER_ADDR", None)


def set_master_port(port: int) -> None:
    global _MASTER_PORT

    if port != _MASTER_PORT:
        _MASTER_PORT = port
        os.environ["MASTER_PORT"] = str(port)
    else:
        raise ValueError(f"Master port {port} is already set")


def get_master_port() -> int:
    assert _MASTER_PORT is not None, "Master port is not yet set"
    return _MASTER_PORT


def clear_master_port() -> None:
    global _MASTER_PORT

    _MASTER_PORT = None
    os.environ.pop("MASTER_PORT", None)


def is_master() -> bool:
    return get_rank() == 0


def is_distributed() -> bool:
    return _INIT_METHOD is not None


def set_init_method(init_method: str) -> None:
    global _INIT_METHOD

    if init_method != _INIT_METHOD:
        os.environ["INIT_METHOD"] = _INIT_METHOD = init_method
    else:
        raise ValueError(f"Init method {init_method} is already set")


def get_init_method() -> str:
    assert _INIT_METHOD is not None, "Init method is not yet set"
    return _INIT_METHOD


def clear_init_method() -> None:
    global _INIT_METHOD

    _INIT_METHOD = None
    os.environ.pop("INIT_METHOD", None)


def set_dist(
    rank: int,
    local_rank: int,
    world_size: int,
    local_world_size: int,
    master_addr: str,
    master_port: int,
    init_method: str,
) -> None:
    set_rank(rank)
    set_local_rank(local_rank)
    set_world_size(world_size)
    set_local_world_size(local_world_size)
    set_master_addr(master_addr)
    set_master_port(master_port)
    set_init_method(init_method)


def clear_dist() -> None:
    clear_rank()
    clear_local_rank()
    clear_world_size()
    clear_local_world_size()
    clear_master_addr()
    clear_master_port()
    clear_init_method()



"""Defines geometry functions."""

import chex
from jax import numpy as jnp
from jaxtyping import Array


def quat_to_euler(quat_4: Array, eps: float = 1e-6) -> Array:
    """Normalizes and converts a quaternion (w, x, y, z) to roll, pitch, yaw.

    Args:
        quat_4: The quaternion to convert, shape (*, 4).
        eps: A small epsilon value to avoid division by zero.

    Returns:
        The roll, pitch, yaw angles with shape (*, 3).
    """
    quat_4 = quat_4 / (jnp.linalg.norm(quat_4, axis=-1, keepdims=True) + eps)
    w, x, y, z = jnp.split(quat_4, 4, axis=-1)

    # Roll (x-axis rotation)
    sinr_cosp = 2.0 * (w * x + y * z)
    cosr_cosp = 1.0 - 2.0 * (x * x + y * y)
    roll = jnp.arctan2(sinr_cosp, cosr_cosp)

    # Pitch (y-axis rotation)
    sinp = 2.0 * (w * y - z * x)

    # Handle edge cases where |sinp| >= 1
    pitch = jnp.where(
        jnp.abs(sinp) >= 1.0,
        jnp.sign(sinp) * jnp.pi / 2.0,  # Use 90 degrees if out of range
        jnp.arcsin(sinp),
    )

    # Yaw (z-axis rotation)
    siny_cosp = 2.0 * (w * z + x * y)
    cosy_cosp = 1.0 - 2.0 * (y * y + z * z)
    yaw = jnp.arctan2(siny_cosp, cosy_cosp)

    return jnp.concatenate([roll, pitch, yaw], axis=-1)


def euler_to_quat(euler_3: Array) -> Array:
    """Converts roll, pitch, yaw angles to a quaternion (w, x, y, z).

    Args:
        euler_3: The roll, pitch, yaw angles, shape (*, 3).

    Returns:
        The quaternion with shape (*, 4).
    """
    # Extract roll, pitch, yaw from input
    roll, pitch, yaw = jnp.split(euler_3, 3, axis=-1)

    # Calculate trigonometric functions for each angle
    cr = jnp.cos(roll * 0.5)
    sr = jnp.sin(roll * 0.5)
    cp = jnp.cos(pitch * 0.5)
    sp = jnp.sin(pitch * 0.5)
    cy = jnp.cos(yaw * 0.5)
    sy = jnp.sin(yaw * 0.5)

    # Calculate quaternion components using the conversion formula
    w = cr * cp * cy + sr * sp * sy
    x = sr * cp * cy - cr * sp * sy
    y = cr * sp * cy + sr * cp * sy
    z = cr * cp * sy - sr * sp * cy

    # Combine into quaternion [w, x, y, z]
    quat = jnp.concatenate([w, x, y, z], axis=-1)

    # Normalize the quaternion
    quat = quat / jnp.linalg.norm(quat, axis=-1, keepdims=True)

    return quat


def get_projected_gravity_vector_from_quat(quat: Array, eps: float = 1e-6) -> Array:
    """Calculates the gravity vector projected onto the local frame given a quaternion orientation.

    Args:
        quat: A quaternion (w,x,y,z) representing the orientation, shape (*, 4).
        eps: A small epsilon value to avoid division by zero.

    Returns:
        A 3D vector representing the gravity in the local frame, shape (*, 3).
    """
    return rotate_vector_by_quat(jnp.array([0, 0, -9.81]), quat, inverse=True, eps=eps)


def rotate_vector_by_quat(vector: Array, quat: Array, inverse: bool = False, eps: float = 1e-6) -> Array:
    """Rotates a vector by a quaternion.

    Args:
        vector: The vector to rotate, shape (*, 3).
        quat: The quaternion to rotate by, shape (*, 4).
        inverse: If True, rotate the vector by the conjugate of the quaternion.
        eps: A small epsilon value to avoid division by zero.

    Returns:
        The rotated vector, shape (*, 3).
    """
    # Normalize quaternion
    quat = quat / (jnp.linalg.norm(quat, axis=-1, keepdims=True) + eps)
    w, x, y, z = jnp.split(quat, 4, axis=-1)

    if inverse:
        x, y, z = -x, -y, -z

    # Extract vector components
    vx, vy, vz = jnp.split(vector, 3, axis=-1)

    # Terms for x component
    xx = (
        w * w * vx
        + 2 * y * w * vz
        - 2 * z * w * vy
        + x * x * vx
        + 2 * y * x * vy
        + 2 * z * x * vz
        - z * z * vx
        - y * y * vx
    )

    # Terms for y component
    yy = (
        2 * x * y * vx
        + y * y * vy
        + 2 * z * y * vz
        + 2 * w * z * vx
        - z * z * vy
        + w * w * vy
        - 2 * w * x * vz
        - x * x * vy
    )

    # Terms for z component
    zz = (
        2 * x * z * vx
        + 2 * y * z * vy
        + z * z * vz
        - 2 * w * y * vx
        + w * w * vz
        + 2 * w * x * vy
        - y * y * vz
        - x * x * vz
    )

    return jnp.concatenate([xx, yy, zz], axis=-1)


def cubic_bezier_interpolation(y_start: Array, y_end: Array, x: Array) -> Array:
    """Cubic bezier interpolation.

    This is a cubic bezier curve that starts at y_start and ends at y_end,
    and is controlled by the parameter x. The curve is defined by the following formula:

    y(x) = y_start + (y_end - y_start) * (x**3 + 3 * (x**2 * (1 - x)))

    Args:
        y_start: The start value, shape (*).
        y_end: The end value, shape (*).
        x: The interpolation parameter, shape (*).

    Returns:
        The interpolated value, shape (*).
    """
    y_diff = y_end - y_start
    bezier = x**3 + 3 * (x**2 * (1 - x))
    return y_start + y_diff * bezier


def quat_to_rotmat(quat: Array, eps: float = 1e-6) -> Array:
    """Converts a quaternion to a rotation matrix.

    Args:
        quat: The quaternion to convert, shape (*, 4).
        eps: A small epsilon value to avoid division by zero.

    Returns:
        The rotation matrix, shape (*, 3, 3).
    """
    quat = quat / (jnp.linalg.norm(quat, axis=-1, keepdims=True) + eps)
    w, x, y, z = jnp.split(quat, 4, axis=-1)

    xx = 1 - 2 * (y * y + z * z)
    xy = 2 * (x * y - z * w)
    xz = 2 * (x * z + y * w)
    yx = 2 * (x * y + z * w)
    yy = 1 - 2 * (x * x + z * z)
    yz = 2 * (y * z - x * w)
    zx = 2 * (x * z - y * w)
    zy = 2 * (y * z + x * w)
    zz = 1 - 2 * (x * x + y * y)

    # Corrected stacking: row-major order
    return jnp.concatenate(
        [
            jnp.concatenate([xx, xy, xz], axis=-1)[..., None, :],
            jnp.concatenate([yx, yy, yz], axis=-1)[..., None, :],
            jnp.concatenate([zx, zy, zz], axis=-1)[..., None, :],
        ],
        axis=-2,
    )


def normalize(v: jnp.ndarray, axis: int = -1, eps: float = 1e-8) -> jnp.ndarray:
    norm = jnp.linalg.norm(v, axis=axis, keepdims=True)
    return v / jnp.clip(norm, min=eps)


def rotation6d_to_rotation_matrix(r6d: jnp.ndarray) -> jnp.ndarray:
    """Convert 6D rotation representation to rotation matrix.

    From https://arxiv.org/pdf/1812.07035, Appendix B

    Args:
        r6d: The 6D rotation representation, shape (*, 6).

    Returns:
        The rotation matrix, shape (*, 3, 3).
    """
    chex.assert_shape(r6d, (..., 6))
    shape = r6d.shape
    flat = r6d.reshape(-1, 6)
    a_1 = flat[:, 0:3]
    a_2 = flat[:, 3:6]

    b_1 = normalize(a_1, axis=-1)

    # Reordered Gram-Schmidt orthonormalization.
    b_3 = normalize(jnp.cross(b_1, a_2), axis=-1)
    b_2 = jnp.cross(b_3, b_1)

    rotation_matrix = jnp.stack([b_1, b_2, b_3], axis=-1)
    return rotation_matrix.reshape(shape[:-1] + (3, 3))


def rotation_matrix_to_rotation6d(rotation_matrix: jnp.ndarray) -> jnp.ndarray:
    """Convert rotation matrix to 6D rotation representation.

    Args:
        rotation_matrix: The rotation matrix, shape (*, 3, 3).

    Returns:
        The 6D rotation representation, shape (*, 6).
    """
    chex.assert_shape(rotation_matrix, (..., 3, 3))
    shape = rotation_matrix.shape
    # Simply concatenate a1 and a2 from SO(3)
    r6d = jnp.concatenate([rotation_matrix[..., 0], rotation_matrix[..., 1]], axis=-1)
    return r6d.reshape(shape[:-2] + (6,))


def quat_mul(q2: Array, q1: Array) -> Array:
    """Multiply two quaternions (supports batching).

    Args:
        q2: Second quaternion (w, x, y, z), shape (..., 4)
        q1: First quaternion (w, x, y, z), shape (..., 4)

    Returns:
        Product quaternion, shape (..., 4)
    """
    w1, x1, y1, z1 = jnp.split(q1, 4, axis=-1)
    w2, x2, y2, z2 = jnp.split(q2, 4, axis=-1)

    w = w2 * w1 - x2 * x1 - y2 * y1 - z2 * z1
    x = w2 * x1 + x2 * w1 + y2 * z1 - z2 * y1
    y = w2 * y1 - x2 * z1 + y2 * w1 + z2 * x1
    z = w2 * z1 + x2 * y1 - y2 * x1 + z2 * w1

    return jnp.concatenate([w, x, y, z], axis=-1)


def rotation_matrix_to_quat(rotation_matrix: Array, eps: float = 1e-6) -> Array:
    """Converts a rotation matrix to a unit quaternion ``(w, x, y, z)``.

    Args:
        rotation_matrix: The rotation matrix, shape ``(*, 3, 3)``.
        eps: A small epsilon value to avoid division by zero when normalising.

    Returns:
        A quaternion with shape ``(*, 4)``.
    """
    chex.assert_shape(rotation_matrix, (..., 3, 3))

    m00 = rotation_matrix[..., 0, 0]
    m01 = rotation_matrix[..., 0, 1]
    m02 = rotation_matrix[..., 0, 2]
    m10 = rotation_matrix[..., 1, 0]
    m11 = rotation_matrix[..., 1, 1]
    m12 = rotation_matrix[..., 1, 2]
    m20 = rotation_matrix[..., 2, 0]
    m21 = rotation_matrix[..., 2, 1]
    m22 = rotation_matrix[..., 2, 2]

    trace = m00 + m11 + m22

    # Case 0: trace is positive
    s0 = jnp.sqrt(jnp.clip(trace + 1.0, min=0.0)) * 2.0  # S = 4 * qw
    w0 = 0.25 * s0
    x0 = (m21 - m12) / jnp.where(s0 < eps, 1.0, s0)
    y0 = (m02 - m20) / jnp.where(s0 < eps, 1.0, s0)
    z0 = (m10 - m01) / jnp.where(s0 < eps, 1.0, s0)

    # Case 1: m00 is the largest diagonal term
    s1 = jnp.sqrt(jnp.clip(1.0 + m00 - m11 - m22, min=0.0)) * 2.0  # S = 4 * qx
    w1 = (m21 - m12) / jnp.where(s1 < eps, 1.0, s1)
    x1 = 0.25 * s1
    y1 = (m01 + m10) / jnp.where(s1 < eps, 1.0, s1)
    z1 = (m02 + m20) / jnp.where(s1 < eps, 1.0, s1)

    # Case 2: m11 is the largest diagonal term
    s2 = jnp.sqrt(jnp.clip(1.0 + m11 - m00 - m22, min=0.0)) * 2.0  # S = 4 * qy
    w2 = (m02 - m20) / jnp.where(s2 < eps, 1.0, s2)
    x2 = (m01 + m10) / jnp.where(s2 < eps, 1.0, s2)
    y2 = 0.25 * s2
    z2 = (m12 + m21) / jnp.where(s2 < eps, 1.0, s2)

    # Case 3: m22 is the largest diagonal term
    s3 = jnp.sqrt(jnp.clip(1.0 + m22 - m00 - m11, min=0.0)) * 2.0  # S = 4 * qz
    w3 = (m10 - m01) / jnp.where(s3 < eps, 1.0, s3)
    x3 = (m02 + m20) / jnp.where(s3 < eps, 1.0, s3)
    y3 = (m12 + m21) / jnp.where(s3 < eps, 1.0, s3)
    z3 = 0.25 * s3

    cond0 = trace > 0.0
    cond1 = (m00 > m11) & (m00 > m22)
    cond2 = m11 > m22

    w = jnp.where(
        cond0,
        w0,
        jnp.where(cond1, w1, jnp.where(cond2, w2, w3)),
    )
    x = jnp.where(
        cond0,
        x0,
        jnp.where(cond1, x1, jnp.where(cond2, x2, x3)),
    )
    y = jnp.where(
        cond0,
        y0,
        jnp.where(cond1, y1, jnp.where(cond2, y2, y3)),
    )
    z = jnp.where(
        cond0,
        z0,
        jnp.where(cond1, z1, jnp.where(cond2, z2, z3)),
    )

    quat = jnp.stack([w, x, y, z], axis=-1)
    quat = quat / (jnp.linalg.norm(quat, axis=-1, keepdims=True) + eps)
    return quat



"""Lets you edit a checkpoint config programmatically."""

import argparse
import difflib
import io
import os
import subprocess
import tarfile
import tempfile
from pathlib import Path

from omegaconf import OmegaConf

from xax.task.mixins.checkpointing import load_ckpt
from xax.utils.text import colored, show_info


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("ckpt_path", type=Path)
    args = parser.parse_args()

    # Loads the config from the checkpoint.
    config = load_ckpt(args.ckpt_path, part="config")
    config_str = OmegaConf.to_yaml(config)

    # Opens the user's preferred editor to edit the config.
    with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False) as f:
        f.write(config_str.encode("utf-8"))
        f.flush()
        subprocess.run([os.environ.get("EDITOR", "vim"), f.name], check=True)

    # Loads the edited config.
    try:
        edited_config = OmegaConf.load(f.name)
        edited_config_str = OmegaConf.to_yaml(edited_config, sort_keys=True)
    finally:
        os.remove(f.name)

    if edited_config_str == config_str:
        show_info("No changes were made to the config.")
        return

    # Diffs the original and edited configs.
    diff = difflib.ndiff(config_str.splitlines(), edited_config_str.splitlines())
    for line in diff:
        if line.startswith("+ "):
            print(colored(line, "light-green"), flush=True)
        elif line.startswith("- "):
            print(colored(line, "light-red"), flush=True)
        elif line.startswith("? "):
            print(colored(line, "light-cyan"), flush=True)

    # Saves the edited config to the checkpoint.
    with tempfile.TemporaryDirectory() as tmp_dir:
        with tarfile.open(args.ckpt_path, "r:gz") as src_tar:
            for member in src_tar.getmembers():
                if member.name != "config":  # Skip the old config file
                    src_tar.extract(member, tmp_dir)

        with tarfile.open(args.ckpt_path, "w:gz") as tar:
            for root, _, files in os.walk(tmp_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    arcname = os.path.relpath(file_path, tmp_dir)
                    tar.add(file_path, arcname=arcname)

            # Add the new config file
            info = tarfile.TarInfo(name="config")
            config_bytes = edited_config_str.encode()
            info.size = len(config_bytes)
            tar.addfile(info, io.BytesIO(config_bytes))


if __name__ == "__main__":
    # python -m xax.cli.edit_config
    main()
